'''

  ponywhoosh
  ~~~~~~~~~~

  Make your database over PonyORM searchable.

  :copyright: (c) 2015-2018 by Jonathan Prieto-Cubides & Felipe Rodriguez.
  :license: MIT (see LICENSE.md)

'''

from __future__     import absolute_import
from __future__     import division
from __future__     import print_function

from .core          import PonyWhoosh
from .utils         import search, full_search, delete_field

__author__  = "Jonathan S. Prieto & Ivan F. Rodriguez"
__all__     = ['PonyWhoosh', 'search', 'full_search', 'delete_field']



'''

  ponywhoosh
  ~~~~~~~~~~

  Makes your database over PonyORM searchable.

  :copyright: (c) 2015-2018 by Jonathan Prieto-Cubides & Felipe Rodriguez.
  :license: MIT (see LICENSE.md)

'''

from __future__   import absolute_import
from __future__   import division
from __future__   import print_function

import re

from collections  import defaultdict
from pony         import orm
from pprint       import pprint
from whoosh       import fields as whoosh_module_fields
from whoosh       import qparser


class Index(object):

  debug = False
  _parameters = {
      'limit'    : 0
    , 'optimize' : False
    , 'reverse'  : False
    , 'scored'   : ''
    , 'sortedby' : ''
  }

  @property
  def whoosh(self):
    return self._whoosh

  @property
  def path(self):
    return self._path

  @property
  def schema(self):
    return self._schema

  @property
  def name(self):
    return self._name

  @property
  def fields(self):
    return self._fields

  def __init__(self, pw):
    """

    Args:
        pw (PonyWhoosh): Initializes of index.
    """
    self._pw   = pw
    self.debug = pw.debug

  def add_field(self, fieldname, fieldspec=whoosh_module_fields.TEXT):
    """Add a field in the index of the model.

    Args:
        fieldname (Text): This parameters register a new field in specified model.
        fieldspec (Name, optional): This option adds various options as were described before.

    Returns:
        TYPE: The new schema after deleted is returned.
    """
    self._whoosh.add_field(fieldname, fieldspec)
    return self._whoosh.schema

  def delete_field(self, field_name):
    """This function deletes one determined field using the command MODEL.pw.delete_field(FIELD)

    Args:
        field_name (string): This argument let you delete some field for some model registered in the index.

    Returns:
        (WhooshSchema): The new schema after deleted is returned.
    """
    self._whoosh.remove_field(field_name.strip())
    return self._whoosh.schema

  def delete_documents(self):
    """Deletes all the  documents using the  pk associated to them.
    """
    pk = str(self._primary_key)
    for doc in self._whoosh.searcher().documents():
      if pk in doc:
        doc_pk = str(doc[pk])
        self._whoosh.delete_by_term(pk, doc_pk)

  def optimize(self):
    """ The index is reindexed, optimizing the run time of searchings and space used. Everytime a document is added a new file would be created, but optimizing would reduce all to just one.

    Returns:
        TYPE: Index optimized.
    """
    self._whoosh.optimize()

  def counts(self):
    """This method counts all the documents contained in  a certain registered model.

    Returns:
        (int): how many documents are indexed of the model index.
    """
    return self._whoosh.doc_count()

  @orm.db_session
  def charge_documents(self):
    """
    This method allow you to charge documents you already have
        in your database. In this way an Index would be created according to
        the model and fields registered.
    """

    doc_count   = self._whoosh.doc_count()
    objs        = orm.count(e for e in self._model)

    field_names = set(self._schema_attrs.keys())
    missings    = set(self._whoosh.schema.names())

    for f in list(field_names - missings):
        self.add_field(f, fields.TEXT(self.kw))

    if doc_count == 0 and objs > 0:
      writer = self._whoosh.writer()
      for obj in orm.select(e for e in self._model):
        attrs = {self._primary_key: obj.get_pk()}
        for f in list(self._schema_attrs.keys()):
          attrs[f] = str(getattr(obj, f))
        writer.add_document(**attrs)
      writer.commit()

  def update_documents(self):
    """It deletes all the documents in the index and charge them again.
    """
    self.delete_documents()
    self.charge_documents()

  @orm.db_session
  def search(self, search_string, **opt):
    """The core function of the package. These are the arguments we consider.


    Args:
        search_string (str): This is what you are looking for in your pony database.
    Optional Args:
        opt: The following opts are available for the search function:
        * add_wildcars(bool): This opt allows you to search no literal queries.
        * fields: This opt let you filter your search result on some model by the fields you want to search.
        * include_entity: This opt let you see not only the result and the register fields but all the instance from the entity of the database.
        * except_fields: Tell the searcher to not look on these fields.
        * use_dict: This option let you activate the dict for the items in the search result, or just view them as a list.
        We implement this because we wanted that the  first field registered of the model would be the most important one.
        As was explained in the IndexView section.
    Returns:
        (dict): A python dictionary with the results for the model.
    """

    prepped_string = self.prep_search_string(
        search_string
      , self.to_bool(opt.get('add_wildcards', False))
      )

    with self._whoosh.searcher() as searcher:
      fields = opt.get('fields', self._schema.names())
      fields = [x for x in fields if len(x) > 0]

      field = opt.get('field', '')
      if len(field) > 0:
        if isinstance(field, str) or isinstance(field, str):
          fields = [field]
        elif isinstance(field, list):
          fields = fields + field

      fields = [x for x in fields if len(x) > 0]

      if len(fields) == 0:
        fields = self._schema.names()

      if len(fields) > 0:
        fields = set(fields) & set(self._schema.names())
        fields = list(fields)

      except_fields = opt.get('except_fields', [])
      except_fields = [x for x in except_fields if len(x) > 0]

      if len(except_fields) > 0:
        fields = list(set(fields) - set(except_fields))

      parser = qparser.MultifieldParser(
          fields, self._whoosh.schema
        , group=opt.get('group', qparser.OrGroup)
        )

      query       = parser.parse(prepped_string)
      search_opts = self.parse_opts_searcher(opt, self._parameters)
      results     = searcher.search(query, terms=True, **search_opts)

      ma = defaultdict(set)
      for f, term in results.matched_terms():
        ma[f].add(term)

      dic = {
          'runtime'       : results.runtime
        , 'cant_results'  : results.estimated_length()
        , 'matched_terms' : {k: list(v) for k, v in list(ma.items())}
        , 'facet_names'   : results.facet_names()
        }

      if dic['cant_results'] == 0 and self.to_bool(opt.get('something', False)):
        opt['add_wildcards'] = True
        opt['something']     = False
        return self.search(search_string, **opt)

      value_results = {}

      for r in results:
        params = {k:r[k] for k in self._primary_key}
        ans = {
            'docnum' : r.docnum
          , 'pk'     : tuple(params.values())
          , 'score'  : r.score
        }

        if self.to_bool(opt.get('include_entity', False)):
          entity     = self._model.get(**params)
          dic_entity = entity.to_dict()

          if opt.get('use_dict', True):
            ans['entity'] = dic_entity
          else:
            fields_missing      = set(dic_entity.keys()) - set(self._fields)
            ans['other_fields'] = [(k, dic_entity[k]) for k in fields_missing]
            ans['entity']       = [(k, dic_entity[k]) for k in self._fields]
          ans['model'] = self._name
        value_results[ans['pk']] = ans

      dic["results"] = list(value_results.values())
      return dic

  def prep_search_string(self, search_string, add_wildcards=False):
    """Prepares search string as a proper whoosh search string.

    Args:
      search_string (str): it prepares the search string and see if
      the lenght is correct.

    Optional Args:
        add_wildcards (bool): It runs a query for inexact queries.

    Raises:
      ValueError: When the search string does not have the appropriate
      lenght. This lenght may be changed in the config options.
    """

    s = search_string.strip()
    try:
      s = str(s)
    except:
      pass
    s = s.replace('*', '')

    if len(s) < self._pw.search_string_min_len:
      raise ValueError('Search string must have at least {} characters'
        .format(self._pw.search_string_min_len))
    if add_wildcards:
      s = '*{0}*'.format(re.sub('[\s]+', '* *', s))
    return s

  def to_bool(self, v):
    if isinstance(v, bool):
      return v
    if isinstance(v, str) or isinstance(v, str):
      return v == 'True' or v == 'true' or v == 't' or v == 'y' or v == 'yes'
    if isinstance(v, int):
      return bool(v)
    return False

  def parse_opts_searcher(self, opts, parameters):
    assert isinstance(opts, dict)
    res = {}
    for k, v in list(opts.items()):
      if k in parameters:
        typevalue = parameters[k]
        if isinstance(typevalue, int):
          if isinstance(v, list):
              res[k] = v[0]
          res[k] = int(v)
        elif isinstance(typevalue, str):
          if isinstance(v, list):
              res[k] = v[0]
          res[k] = str(v)
        elif isinstance(typevalue, bool):
          if isinstance(v, list):
              res[k] = v[0]
          res[k] = self.to_bool(v)
    return res



'''

  ponywhoosh
  ~~~~~~~~~~

  Make your database over PonyORM searchable.

  :copyright: (c) 2015-2018 by Jonathan Prieto-Cubides & Felipe Rodriguez.
  :license: MIT (see LICENSE.md)

'''

from __future__              import absolute_import
from __future__              import division
from __future__              import print_function

import os
import re
import sys
import whoosh

from collections             import defaultdict
from .index                  import Index  as PonyWhooshIndex
from pony                    import orm
from pony.orm.serialization  import to_dict
from pprint                  import pprint
from whoosh                  import fields as whoosh_module_fields
from whoosh                  import index  as whoosh_module_index
from whoosh                  import qparser

__all__   = ['PonyWhoosh']
basedir   = os.path.abspath(os.path.dirname(__file__))


class PonyWhoosh(object):

  """A top level class that allows to register indexes and performing searches.

  Attributes:
      * debug (bool): print some messages useful for debugging
      * indexes_path (str): this is the name where the folder of the indexes
        are going to be stored.
      * search_string_min_len (int): let you config the minimun string value
        possible to perform search.
      * writer_timeout (int): a time constraint for running a search process.
  """

  debug                 = False
  indexes_path          = None
  search_string_min_len = 2
  writer_timeout        = 4

  _indexes              = {}
  _entities             = {}

  # def __init__(self, indexes_path=None):

  #   if indexes_path is None or indexes_path == "":
  #     indexes_path = "indexes"
  #   if not os.path.exists(self.indexes_path):
  #     os.makedirs(self.indexes_path)

  def delete_indexes(self):
    """This set to empty all the indexes registered.
    """
    self._indexes = {}

  def indexes(self):
    """Access a list of the current indexes registered

    Returns:
        (list): the indexes stored.
    """
    return [v for k, v in list(self._indexes.items())]

  def create_index(self, index):
    """Creates and opens index folder for given index.
    If the index already exists, it just opens it, otherwise it creates it first.
    """

    index._path = os.path.join(self.indexes_path, index._name)

    if whoosh.index.exists_in(index._path):
      _whoosh = whoosh.index.open_dir(index._path)
    elif not os.path.exists(index._path):
      os.makedirs(index._path)
      _whoosh = whoosh.index.create_in(index._path, index._schema)
    index._whoosh = _whoosh

  def register_index(self, index):
    """Registers a given index:

    * Creates and opens an index for it (if it doesn't exist yet)
    * Sets some default values on it (unless they're already set)

    Args:
        index (PonyWhoosh.Index): An instance of PonyWhoosh.Index class
    """

    self._indexes[index._name] = index
    self.create_index(index)
    return index

  def register_model(self, *fields, **kw):
    """Registers a single model for fulltext search. This basically creates
    a simple PonyWhoosh.Index for the model and calls self.register_index on it.

    Args:
        *fields: all the fields indexed from the model.
        **kw: The options for each field, sortedby, stored ...
    """

    index         = PonyWhooshIndex(pw=self)
    index._kw     = kw
    index._fields = fields

    def inner(model):
      """This look for the types of each field registered in the index,
      whether if it is Numeric, datetime, boolean or just text.

      Args:
          model (PonyORM.Entity): A model defined on the database with PonyORM

      Returns:
          model (PonyORM.Entity): A model modified for handle the appropriate
          search methods.
      """

      index._name = model._table_
      if not index._name:
        index._name  = model.__name__

      self._entities[index._name]     = model
      index._schema_attrs             = {}
      index._primary_key_is_composite = model._pk_is_composite_
      index._primary_key              = [f.name for f in model._pk_attrs_]
      index._primary_key_type         = 'list'
      type_attribute                  = {}

      for field in model._attrs_:
        if field.is_relation:
          continue

        assert hasattr(field, "name") and hasattr(field, "py_type")

        fname = field.name
        if hasattr(field.name, "__name__"):
            fname = field.name.__name__

        stored = kw.get("stored", False)
        if fname in index._primary_key:
            kw["stored"] = True
        # we're not supporting this kind of data
        ftype = field.py_type.__name__
        if ftype in ['date', 'datetime', 'datetime.date']:
            kw["stored"] = stored
            continue

        fwhoosh = fwhoosh = whoosh.fields.TEXT(**kw)

        if field == model._pk_:
            index._primary_key_type = ftype
            fwhoosh = whoosh.fields.ID(stored=True, unique=True)

        if fname in index._fields:
          if not field.is_string:
            if ftype in ['int', 'float']:
              fwhoosh = whoosh.fields.NUMERIC(**kw)
            elif ftype == 'bool':
              fwhoosh = whoosh.fields.BOOLEAN(stored=True)

        type_attribute[fname]      = ftype
        index._schema_attrs[fname] = fwhoosh
        kw["stored"]               = stored

      index._schema = whoosh.fields.Schema(**index._schema_attrs)

      self.register_index(index)

      def _middle_save_(obj, status):
        """A middle-in-middle method to intercept CRUD operations from PonyORM
        over the current object model to update the appropriate whoosh index.

        Args:
            obj (EntityInstance): An instance of a current model.
            status (str): Type of transaction on the database. A CRUD operation.

        Returns:
            obj (EntityInstance): The same object as the input.
        """

        writer   = index._whoosh.writer(timeout=self.writer_timeout)
        dict_obj = obj.to_dict()

        def dumps(v):
          if sys.version_info[0] < 3:
            if isinstance(v, int):
              return unicode(v)
            if isinstance(v, float):
              return '%.9f' % v
            return unicode(v)
          else:
            if isinstance(v, int):
              return str(v)
            if isinstance(v, float):
              return int(float(v))
            return str(v)

        attrs = {}
        if sys.version_info[0] < 3:
          for k, v in dict_obj.iteritems():
            if k in index._schema_attrs.keys():
              attrs[k] = dumps(v)
        else:
          for k, v in dict_obj.items():
            if k in list(index._schema_attrs.keys()):
              attrs[k] = dumps(v)

        if status == 'inserted':
          writer.add_document(**attrs)
        elif status == 'updated':
          writer.update_document(**attrs)
        elif status in set(['marked_to_delete', 'deleted', 'cancelled']):
          writer.delete_by_term(primary, attrs[primary])

        writer.commit()
        return obj._after_save_

      index._model       = model
      model._after_save_ = _middle_save_
      model._pw_index_   = index
      model.search       =  model._pw_index_.search
      return model
    return inner

  @orm.db_session
  def search(self, *arg, **kw):
    """A full search function. This allows you to search expression
    using the following arguments.

    Arg:
        query (str): The search string expression.

    Optional Args:
        - include_entity (bool): include in each result the entity values associated of the fields stored.
        - add_wildcards (bool): set it if you want to consider matches that have prefix or suffixes the query.
        - something (bool): set `add_willcards` in case of none results for the query.
        - fields (list): specified the fields names that you want to consider.
        - except_fields (list): specified the fields names to not consider in the search.
        - models (list): a list of name of model to search or even the models from the database.

    Returns:
        (dict): A python dictionary with the results.
    """
    output = {
        'cant_results'  : 0
      , 'matched_terms' : defaultdict(set)
      , 'results'       : {}
      , 'runtime'       : 0
    }

    indexes = self.indexes()

    models = kw.get('models', list(self._entities.values()))
    if sys.version_info[0] < 3:
      models = [self._entities.get(model, None) if isinstance(model, str)
        or isinstance(model, unicode) else model for model in models]
      models = filter(lambda x: x is not None, models)
    else:
      models = [self._entities.get(model, None) if isinstance(model, str)
        or isinstance(model, str) else model for model in models]
      models = [x for x in models if x is not None]


    if models == [] or not models:
      models = list(self._entities.values())

    if self.debug:
      print("SEARCHING ON MODELS -> ", models)

    indexes = [m._pw_index_ for m in models if hasattr(m, '_pw_index_')]

    if indexes == []:
      return output

    runtime, cant = 0, 0
    ma            = defaultdict(set)

    for index in indexes:
      res     = index.search(*arg, **kw)
      runtime += res['runtime']
      cant    += res['cant_results']
      if res['cant_results'] > 0:
        output['results'][index._name] = {
            'items'         : res['results']
          , 'matched_terms' : res['matched_terms']
        }
        for k, ts in list(res['matched_terms'].items()):
          for t in ts:
            ma[k].add(t)

    output['cant_results']  = cant
    output['matched_terms'] = {k: list(v) for k, v in list(ma.items())}
    output['runtime']       = runtime
    return output



'''

  ponywhoosh
  ~~~~~~~~~~

  Makes your database over PonyORM searchable.

  :copyright: (c) 2015-2018 by Jonathan Prieto-Cubides & Felipe Rodriguez.
  :license: MIT (see LICENSE.md)

'''
from __future__  import absolute_import
from __future__  import division
from __future__  import print_function

from pony        import orm

@orm.db_session
def search(model, *arg, **kw):
  """Ponywhoosh function to perform searches on specific models.
  It takes up three arguments:

  Args:
    model (Model) or (str): Where you want to search, in which table.
    query (str): the string expression to search.

  Optional Args:
    - include_entity (bool): include in each result the entity values
      associatedof the fields stored.
    - add_wildcards (bool): set it if you want to consider matches that
      have prefix or suffixes the query.
    - something (bool): set `add_willcards` in case of none results
      for the query.
    - fields (list): specified the fields names that you want to consider.
    - except_fields (list): specified the fields names to not consider in
      the search.
    - models (list): a list of name of model to search or even the models
      from the database.

  Returns:
    (dict): A python dictionary with the results for the model.
  """
  return model._pw_index_.search(*arg, **kw)


@orm.db_session
def delete_field(model, *arg):
  """ It deletes an specific field stored in the index.

  Args:
    model (Model): The model from where you want to delete the specific field(s).
    *arg: field(s) of the model specified that will be removed.

  Returns:
    (list): a schemma of the model with the fields removed.
  """
  return model._pw_index_.delete_field(*arg)


def full_search(pw, *arg, **kw):
  """ This function search in every model registered.
  And portrays the result in a dictionary where the keys are the models.

  Args:
    pw (PonyWhoosh): This is where all the indexes are stored.
    An instance of PonyWhoosh class.
    query (str): the string expression to search.

  Optional Args:
    - include_entity (bool): include in each result the entity
      values associated of the fields stored.
    - add_wildcards (bool): set it if you want to consider matches
      that have prefix or suffixes the query.
    - something (bool): set `add_willcards` in case of none results
      for the query.
    - fields (list): specified the fields names that you want to consider.
    - except_fields (list): specified the fields names to not consider in
      the search.
    - models (list): a list of name of model to search or even the models
      from the database.

  Returns:
    (dict): A python dictionary with the results for the model.
  """
  return pw.search(*arg, **kw)



'''

  ponywhoosh
  ~~~~~~~~~~

  Your database now searchable.

  :copyright: (c) 2015-2018 by Jonathan S. Prieto & Ivan Felipe Rodriguez.
  :license: BSD (see LICENSE.md)

'''

import io
import os
import re

from glob       import glob
from os.path    import basename, dirname, join, relpath, splitext
from setuptools import find_packages, setup


def read(*names, **kwargs):
  return io.open(
      join(dirname(__file__), *names),
      encoding=kwargs.get('encoding', 'utf8')
  ).read()

setup(
    name='ponywhoosh'
  , version="1.7.8"
  , url='https://github.com/jonaprieto/ponywhoosh'
  , license='MIT'
  , author='Jonathan Prieto-Cubides & Felipe Rodriguez'
  , author_email='jprieto9@eafit.edu.co'
  , description='Your database now searchable. The backend behind the Flask-PonyWhoosh.'
  , long_description='%s' % ( read('README.rst') )
  , packages=find_packages()
  , zip_safe=False
  , include_package_data=True
  , platforms='any'
  , keywords=
    [ 'elastic'
    , 'engine'
    , 'flask'
    , 'flask-sqlalchemy'
    , 'flask-whooshalchemy'
    , 'mysql'
    , 'pony'
    , 'ponyorm'
    , 'ponywhoosh'
    , 'search'
    , 'searchable'
    , 'sqlite3'
    , 'whoosh'
    ]
  , install_requires=['pony', 'whoosh']
  , classifiers=[
      'Environment :: Web Environment'
    , 'Intended Audience :: Developers'
    , 'License :: OSI Approved :: MIT License'
    , 'Operating System :: OS Independent'
    , 'Programming Language :: Python'
    , 'Topic :: Internet :: WWW/HTTP :: Dynamic Content'
    , 'Topic :: Software Development :: Libraries :: Python Modules'
  ]
)

# pandoc --from=rst --to=rst --output=README.rst README.rst
# Pasos para subir a pypi
# git tag v...
# python setup.py register -r pypi
# python setup.py sdist upload -r pypi
